from pathlib import Path
from typing import Set, Union

from poetry.console.commands.command import Command
from poetry.factory import Factory
from polylith import info, project, repo, workspace
from polylith.libs import report


class LibsCommand(Command):
    name = "poly libs"
    description = "Show third-party libraries used in the workspace."

    def find_third_party_libs(self, path: Union[Path, None]) -> Set:
        project_poetry = Factory().create_poetry(path) if path else self.poetry

        if not project_poetry.locker.is_locked():
            raise ValueError("poetry.lock not found. Run `poetry lock` to create it.")

        packages = project_poetry.locker.locked_repository().packages

        return {p.name for p in packages}

    def print_report(self, root: Path, ns: str, data: dict) -> bool:
        name = data["name"]
        path = data["path"]

        brick_imports = report.get_third_party_imports(root, ns, data)

        report.print_libs_summary(brick_imports, data)
        report.print_libs_in_bricks(brick_imports)

        try:
            third_party_libs = self.find_third_party_libs(path)

            return report.print_missing_installed_libs(
                brick_imports, third_party_libs, name
            )
        except ValueError as e:
            self.line_error(f"{name}: <error>{e}</error>")
            return False

    def handle(self) -> int:
        root = repo.find_workspace_root(Path.cwd())

        if not root:
            raise ValueError(
                "Didn't find the workspace root. Expected to find a workspace.toml file."
            )

        ns = workspace.parser.get_namespace_from_config(root)

        projects_data = info.get_projects_data(root, ns)

        if self.option("directory"):
            project_name = project.get_project_name(self.poetry.pyproject.data)

            data = next((p for p in projects_data if p["name"] == project_name), None)

            if not data:
                raise ValueError(f"Didn't find project in {self.option('directory')}")

            res = self.print_report(root, ns, data)
            result_code = 0 if res else 1
        else:
            results = {self.print_report(root, ns, data) for data in projects_data}

            result_code = 0 if all(results) else 1

        return result_code



from pathlib import Path
from typing import Set

from polylith import imports, libs, workspace
from polylith.check import grouping
from polylith.reporting import theme
from rich.console import Console


def print_missing_deps(brick_imports: dict, deps: Set[str], project_name: str) -> bool:
    diff = libs.report.calculate_diff(brick_imports, deps)

    if not diff:
        return True

    console = Console(theme=theme.poly_theme)

    missing = ", ".join(sorted(diff))

    console.print(f":thinking_face: Cannot locate {missing} in {project_name}")
    return False


def print_report(
    root: Path, ns: str, project_data: dict, third_party_libs: Set
) -> bool:
    name = project_data["name"]

    bases = {b for b in project_data.get("bases", [])}
    components = {c for c in project_data.get("components", [])}

    bases_paths = workspace.paths.collect_bases_paths(root, ns, bases)
    components_paths = workspace.paths.collect_components_paths(root, ns, components)

    all_imports_in_bases = imports.fetch_all_imports(bases_paths)
    all_imports_in_components = imports.fetch_all_imports(components_paths)

    brick_imports = {
        "bases": grouping.extract_brick_imports(all_imports_in_bases, ns),
        "components": grouping.extract_brick_imports(all_imports_in_components, ns),
    }

    third_party_imports = {
        "bases": libs.extract_third_party_imports(all_imports_in_bases, ns),
        "components": libs.extract_third_party_imports(all_imports_in_components, ns),
    }

    packages = set().union(bases, components)

    brick_result = print_missing_deps(brick_imports, packages, name)
    libs_result = print_missing_deps(third_party_imports, third_party_libs, name)

    return all([brick_result, libs_result])



from typing import List

from polylith.reporting import theme
from rich import box
from rich.columns import Columns
from rich.console import Console
from rich.padding import Padding
from rich.table import Table


def brick_status(brick, bricks) -> str:
    status = ":gear:" if brick in bricks else "-"

    return f"[data]{status}[/]"


def print_diff_details(
    projects_data: List[dict], bases: List[str], components: List[str]
) -> None:

    if not bases and not components:
        return

    console = Console(theme=theme.poly_theme)
    table = Table(box=box.SIMPLE_HEAD)
    table.add_column("[data]changed brick[/]")

    proj_cols = [f"[proj]{project['name']}[/]" for project in projects_data]
    table.add_column(Columns(proj_cols, align="center", expand=True))

    for brick in sorted(components):
        cols = [brick_status(brick, p.get("components")) for p in projects_data]
        table.add_row(f"[comp]{brick}[/]", Columns(cols, align="center", expand=True))

    for brick in sorted(bases):
        cols = [brick_status(brick, p.get("bases")) for p in projects_data]
        table.add_row(f"[base]{brick}[/]", Columns(cols, align="center", expand=True))

    console.print(table, overflow="ellipsis")


def print_detected_changes_in_projects(projects: List[str]) -> None:
    if not projects:
        return

    console = Console(theme=theme.poly_theme)

    for project in sorted(projects):
        console.print(f"[data]:gear: Changes found in [/][proj]{project}[/]")


def print_diff_summary(tag: str, bases: List[str], components: List[str]) -> None:
    console = Console(theme=theme.poly_theme)

    console.print(Padding(f"[data]Diff: based on the {tag} tag[/]", (1, 0, 1, 0)))

    if not bases and not components:
        console.print("[data]No brick changes found.[/]")
        return

    if components:
        console.print(f"[comp]Changed components[/]: [data]{len(components)}[/]")

    if bases:
        console.print(f"[base]Changed bases[/]: [data]{len(bases)}[/]")


def _changed_projects(
    projects_data: List[dict], brick_type: str, bricks: List[str]
) -> set:
    res = {
        p["path"].name: set(p.get(brick_type, [])).intersection(bricks)
        for p in projects_data
    }

    return {k for k, v in res.items() if v}


def print_short_diff(
    projects_data: List[dict],
    projects: List[str],
    bases: List[str],
    components: List[str],
) -> None:

    a = _changed_projects(projects_data, "components", components)
    b = _changed_projects(projects_data, "bases", bases)
    c = set(projects)

    res = {*a, *b, *c}

    console = Console(theme=theme.poly_theme)
    console.print(",".join(res))



from pathlib import Path
from typing import List, Set

from polylith import check, imports, info, workspace


def get_brick_imports(root: Path, ns: str, project_data: dict) -> dict:
    bases = {b for b in project_data.get("bases", [])}
    components = {c for c in project_data.get("components", [])}

    bases_paths = workspace.paths.collect_bases_paths(root, ns, bases)
    components_paths = workspace.paths.collect_components_paths(root, ns, components)

    all_imports_in_bases = imports.fetch_all_imports(bases_paths)
    all_imports_in_components = imports.fetch_all_imports(components_paths)

    return {
        "bases": check.grouping.extract_brick_imports(all_imports_in_bases, ns),
        "components": check.grouping.extract_brick_imports(
            all_imports_in_components, ns
        ),
    }


def diff(known_bricks: Set[str], bases: List[str], components: List[str]) -> Set[str]:
    bricks = set().union(bases, components)

    return known_bricks.difference(bricks)


def imports_diff(imports: dict, bases: List, components: List) -> Set[str]:
    flattened_bases = set().union(*imports["bases"].values())
    flattened_components = set().union(*imports["components"].values())

    flattened_imports = set().union(flattened_bases, flattened_components)

    return diff(flattened_imports, bases, components)


def calculate_diff(
    root: Path,
    namespace: str,
    project_data: dict,
    workspace_data: dict,
) -> dict:
    imports = get_brick_imports(root, namespace, project_data)

    all_bases = workspace_data["bases"]
    all_components = workspace_data["components"]

    bases = project_data["bases"]
    components = project_data["components"]

    is_project = info.is_project(project_data)

    if is_project:
        brick_diff = imports_diff(imports, bases, components)
    else:
        all_bricks = set().union(all_bases, all_components)
        brick_diff = diff(all_bricks, bases, components)

    return {
        "name": project_data["name"],
        "path": project_data["path"],
        "is_project": is_project,
        "bases": {b for b in brick_diff if b in all_bases},
        "components": {b for b in brick_diff if b in all_components},
    }
